""" Actually making requests to the API """

import json
import requests
import urllib

import ot_api

headers = {
  "Opentrons-Version": "3",
}

def get_base() -> str:
  if ot_api.HOST is None:
    raise RuntimeError("ot_api host is not set, use ot_api.set_host('x.x.x.x')")
  base = f'http://{ot_api.HOST}:{ot_api.PORT}'
  return base

def get_url(path):
  return urllib.parse.urljoin(get_base(), path)

def _return_resp(resp: requests.Response) -> dict:
  if resp.status_code not in range(200, 300):
    raise Exception(resp.text)
  return json.loads(resp.text)

def get(path: str) -> dict:
  url = get_url(path)
  resp = requests.get(url, headers=headers)
  return _return_resp(resp)

def post(path: str, data=None) -> dict:
  url = get_url(path)
  if data is not None:
    data = json.dumps(data)
    headers["Content-Type"] = "application/json"
  resp = requests.post(url, headers=headers, data=data)
  return _return_resp(resp)

def delete(path):
  url = get_url(path)
  resp = requests.delete(url, headers=headers)
  return _return_resp(resp)



""" Somewhat nicer wrapper around ot_api.runs for labware related things """

from ot_api.decorators import request_with_run_id, command
import ot_api.requestor
import ot_api.runs


@request_with_run_id
def define(labware_def, run_id=None):
  data = dict(data=labware_def)
  return ot_api.requestor.post(f"/runs/{run_id}/labware_definitions", data)

@request_with_run_id
def undefine(labware_def_id, run_id=None):
  return ot_api.requestor.delete(f"/runs/{run_id}/labware_definitions/{labware_def_id}")

@command
def add(load_name, namespace, version, slot: int, run_id: str = None, labware_id=None, display_name=None):
  assert slot in range(1, 13)
  data = {
    "location": {
      "slotName": str(slot),
    },
    "loadName": load_name,
    "namespace": namespace,
    "version": version,
    "labwareId": labware_id,
    "displayName": display_name,
  }
  return ot_api.runs.enqueue_command("loadLabware", data, intent="setup", run_id=run_id)



""" Somewhat nicer wrapper around ot_api.runs for lh related things """

from typing import Optional, Tuple

from ot_api.decorators import command, request_with_run_id
import ot_api.requestor as requestor
import ot_api.runs

def load_pipette(pipette_name, mount, run_id: Optional[str] = None):
  assert mount in ["left", "right"]

  @command
  def _load_pipette(run_id=None):
    data = {
      "pipetteName": pipette_name,
      "mount": mount,
    }
    return ot_api.runs.enqueue_command("loadPipette", params=data, intent="setup", run_id=run_id)

  resp = _load_pipette(run_id=run_id)
  return resp["data"]["result"]

@request_with_run_id
def add_mounted_pipettes(run_id=None) -> Tuple[Optional[dict], Optional[dict]]: # TODO: TypedDict
  mounted_pipettes = requestor.get("/pipettes")

  left_pipette = mounted_pipettes["left"]
  if left_pipette["name"] is not None:
    left = load_pipette(left_pipette["name"], "left", run_id=run_id)
    left["name"] = left_pipette["name"]
  else:
    left = None

  right_pipette = mounted_pipettes["right"]
  if right_pipette["name"] is not None:
    right = load_pipette(right_pipette["name"], "right", run_id=run_id)
    right["name"] = right_pipette["name"]
  else:
    right = None

  return left, right

@command
def pick_up_tip(
  labware_id: str,
  well_name: str,
  pipette_id: str,
  offset_x: float = 0,
  offset_y: float = 0,
  offset_z: float = 0,
  run_id: Optional[str]=None
):
  params = {
    "labwareId": labware_id,
    "wellName": well_name,
    "wellLocation": {
      "origin": "top",
      "offset": {
        "x": offset_x,
        "y": offset_y,
        "z": offset_z
      }
    },
    "pipetteId": pipette_id
  }
  return ot_api.runs.enqueue_command("pickUpTip", params, intent="setup", run_id=run_id)

@command
def drop_tip(
  labware_id: str,
  well_name: str,
  pipette_id: str,
  offset_x: float = 0,
  offset_y: float = 0,
  offset_z: float = 0,
  run_id: Optional[str]=None
):
  params = {
    "labwareId": labware_id,
    "wellName": well_name,
    "wellLocation": {
      "origin": "top",
      "offset": {
        "x": offset_x,
        "y": offset_y,
        "z": offset_z
      }
    },
    "pipetteId": pipette_id
  }

  return ot_api.runs.enqueue_command("dropTip", params, intent="setup", run_id=run_id)

@command
def aspirate(
  labware_id: str,
  well_name: str,
  volume: float,
  flow_rate: float,
  pipette_id,
  run_id: Optional[str]=None,
  offset_x: float = 0,
  offset_y: float = 0,
  offset_z: float = 0,
):
  params = {
    "labwareId": labware_id,
    "wellName": well_name,
    "wellLocation": {
      "origin": "top",
      "offset": {
        "x": offset_x,
        "y": offset_y,
        "z": offset_z
      },
    },
    "flowRate": flow_rate,
    "volume": volume,
    "pipetteId": pipette_id
  }

  return ot_api.runs.enqueue_command("aspirate", params, intent="setup", run_id=run_id)

@command
def dispense(
  labware_id: str,
  well_name: str,
  volume: float,
  flow_rate: float,
  pipette_id,
  run_id: Optional[str]=None,
  offset_x: float = 0,
  offset_y: float = 0,
  offset_z: float = 0,
):
  params = {
    "labwareId": labware_id,
    "wellName": well_name,
    "wellLocation": {
      "origin": "top",
      "offset": {
        "x": offset_x,
        "y": offset_y,
        "z": offset_z
      },
    },
    "flowRate": flow_rate,
    "volume": volume,
    "pipetteId": pipette_id
  }

  return ot_api.runs.enqueue_command("dispense", params, intent="setup", run_id=run_id)



""" Wrapper around the /runs endpoints"""

from typing import Optional

from ot_api.decorators import request_with_run_id
import ot_api.requestor as requestor

def create() -> str:
  data = requestor.post("/runs")
  return data["data"]["id"]

def get_all() -> list:
  data = requestor.get("/runs")
  return data["data"]

def get_run(run_id: str) -> dict:
  data = requestor.get(f"/runs/{run_id}")
  return data["data"]

@request_with_run_id
def get_current_run(run_id) -> str:
  return get_run(run_id)

@request_with_run_id
def get_command(command_id: str, run_id: Optional[str]):
  data = requestor.get(f"/runs/{run_id}/commands/{command_id}")
  return data

@request_with_run_id
def enqueue_command(command, params, intent, run_id):
  assert intent in ["setup", "protocol"]
  command = {
    "data": {
      "commandType": command,
      "params": params,
      "intent": intent
    }
  }
  resp = requestor.post(f"/runs/{run_id}/commands", command)
  return resp["data"]["id"]



""" Decorators for the ot_api functions """

import datetime
import functools

import opentrons.protocol_engine.errors as ot_errors

import ot_api


def request_with_run_id(f):
  """ get run_id from param, if given, otherwise __init__, if given, otherwise raise error """

  @functools.wraps(f)
  def wrapper(*args, **kwargs):
    if "run_id" not in kwargs:
      run_id = ot_api.run_id
      if run_id is None:
        raise TypeError("No run_id given. Please pass run_id as a parameter or use ot_api.set_run_id()")
      kwargs["run_id"] = run_id

    try:
      return f(*args, **kwargs)
    except TypeError as e:
      if "run_id" in str(e):
        raise TypeError("Error calling function. Did you not pass run_id as a kwarg?")
      raise e

  return wrapper


def command(f, timeout=30):
  """ Decorator for commands. Uses request_with_run_id. Waits for success or failure, potentially raising an error. """

  def get_ot_error(name) -> ot_errors.ProtocolEngineError:
    return getattr(ot_errors, name)
  
  @request_with_run_id
  def wrapper(*args, **kwargs):
    command_id = f(*args, **kwargs)

    end = datetime.datetime.now() + datetime.timedelta(seconds=timeout)
    while datetime.datetime.now() < end:
      result = ot_api.runs.get_command(command_id, run_id=kwargs["run_id"])
    
      if result["data"]["status"] == "failed":
        error_data = result["data"]["error"]
        error_class = get_ot_error(error_data["errorType"])
        raise error_class(error_data["detail"])
      elif result["data"]["status"] in ["queued", "running"]:
        continue
 
      return result
    
    raise RuntimeError("Command timed out")

  return wrapper



HOST = None
PORT = 31950

run_id = None

def set_host(new_host):
  global HOST
  HOST = new_host

def set_port(new_port):
  global PORT
  PORT = new_port

def set_run(new_run_id):
  global run_id
  run_id = new_run_id

import ot_api.health as health
import ot_api.labware as labware
import ot_api.lh as lh
import ot_api.runs as runs



from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
  long_description = f.read()

setup(
    name='opentrons http api client',
    version="0.1.2",
    python_requires=">=3.7",
    packages=find_packages(exclude="testing"),
    description='A Python library for interacting with the Opentrons HTTP API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
      "opentrons",
      "requests",
    ],
    url='https://github.com/rickwierenga/opentrons-python-api.git',
)
